module MyBuffer (Buf, new, put, get) where

import Control.Concurrent.STM
import Control.Concurrent
import Control.Monad(forever)
import Data.Time(getCurrentTime)
import System.Random(randomRIO)

{-
Remember:
check :: Bool -> STM a
check b | b = return ()
        | not b = retry
        
-}

newtype Buf a = B (TVar [a])

new :: STM (Buf a)
new = do tv <- newTVar []
         return $ B tv

put :: Buf a -> a -> STM ()
put (B tv) x = do xs <- readTVar tv
                  writeTVar tv (xs ++ [x])

get :: Buf a -> STM a
get (B tv) = do xs <- readTVar tv
                case xs of
                  [] -> retry
                  (y:xs') -> do writeTVar tv xs'
                                return y


useBuffer :: IO ()
useBuffer = do
  b <- atomically $ new 
  forkIO $ forever $ do
    n<- randomRIO(1,5)
    threadDelay (n*10^6)
    t <- getCurrentTime 
    mapM_ (\x-> atomically $ put b $ show x) (replicate n t)
  forever $ do x <- atomically $ get b
               putStrLn $ x

