module Philosophers where

-- Uses code from http://rosettacode.org/wiki/Dining_philosophers#Haskell 
 
import Control.Monad
import Control.Concurrent
import Control.Concurrent.STM
import System.Random
 
-- A fork state:  either on the table (down), or taken. 
data FS = Down | Taken deriving Eq

-- A fork has a number (for display purposes) and a state
data Fork = Fork { fid :: Int, tvar :: TVar FS }

newFork :: Int -> IO Fork
newFork i = atomically $ do 
  f<- newTVar Down
  return $ Fork i f
               
-- Take a fork
takeFork :: Fork -> STM ()
takeFork (Fork _ f) = do 
  s<- readTVar f
  when (s == Taken) retry 
  writeTVar f Taken 
 
-- Release a fork
releaseFork :: Fork -> STM ()
releaseFork (Fork _ f) = writeTVar f Down
 
philosopher :: String -> Fork-> Fork -> IO ()
philosopher name left right = forever $ do
  putStrLn $ name ++ " is thinking."
  delay <- randomRIO (1, 10)
  threadDelay (delay * 1000000)

  putStrLn $ name ++ " is hungry."
  -- Take the forks 
  atomically $ do  {takeFork left; takeFork right}
  putStrLn $ name ++ " got forks " ++ show (fid left) ++ " and " ++ show (fid right) ++ " and is now eating."
  delay <- randomRIO (1,10)
  threadDelay (delay * 1000000) -- 1 to 10 seconds. 

  putStrLn $ name ++ " is done eating. Going back to thinking." 
  -- Put down the forks
  atomically $ do {releaseFork left; releaseFork right}
 

philosophers :: [String]
philosophers = ["Aristotle", "Kant", "Spinoza", "Marx", "Russel"]
 
named :: IO ()
named = do
  forks <- mapM newFork [1..5]
  forM [0..4] $ \i-> 
     forkIO (philosopher (philosophers !! i) 
                         (forks !! i)
                         (forks !! ((i+1) `mod` 4))) 
  getLine
  return ()

many :: Int-> IO ()
many n = do
  forks <- mapM newFork [1..n]
  forM [0..n-1] $ \i-> 
     forkIO (philosopher ("Philosopher #"++ show (i+1))
                         (forks !! i)
                         (forks !! ((i+1) `mod` n))) 
  getLine
  return ()
